==> lmtpd-0.9.19.tar.gz: lmtpd-0.9.19/filter.c <==
/*
   Copyright 2000, 2001, 2002, 2003 Laurent Wacrenier

   This file is part of PLL lmtpd.

   PLL lmtpd is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   PLL lmtpd is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with PLL lmtpd; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

static char const rcsid[] =
"$Id: filter.c,v 1.65 2004/04/20 13:16:53 lwa Exp $";

#include "config.h"

#include <sys/types.h>
#include <syslog.h>

#include <limits.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <strings.h>
#include <string.h>

#include "lmtp.h"

#define BUFMAX 1024

#if !HAVE_ISBLANK
int isblank(int c) {
  return c == ' ' || c == '\t';
}
#endif

#define skipblanks(xbuf) while(isblank((int)*(xbuf))) (xbuf)++

#define CMD(x) { #x, sizeof(#x)-1, cmd_##x, }
#define CMD_END { NULL, 0, NULL, }

struct cmd cmdlist[] = {
  /* { "deliver", sizeof("deliver")-1, cmd_deliver}, */
  CMD(deliver),
  CMD(drop),
  CMD(bounce),
  CMD(forward),
  CMD(forward2),
  CMD(skipto),
  CMD(score),
  CMD(trash),
  CMD(vacation),
  CMD(logrotate),
  CMD(trace),
  CMD(timezone),
  CMD(mod),
  CMD(add),
  CMD(blacklist),
  CMD(extension),
  CMD(stow),
  CMD(log),
  CMD(syslog),
  CMD(deliver2),
  CMD(default),
  CMD(antivirus),
  CMD(spamassassin),
  CMD(header_set),
  CMD(archive),
  CMD(archive2),
  CMD(moved),
  CMD(pipe),
  CMD(debug),
  CMD(stop),
  CMD_END,
};

struct cmd mod_cmdlist[] = {
  CMD(score),
  CMD(message),
  CMD(ok),
  CMD(skipto),
  CMD_END,
};

#define LOGTYPEMODE_SET   '='
#define LOGTYPEMODE_PLUS  '+'
#define LOGTYPEMODE_MINUS '-'

struct logtype {
  char *key;
  unsigned char plus;
  unsigned char minus;
} logtypes[] = {
  { "on",  ACTION_LOG_ALL, ACTION_LOG_ALL & ~ACTION_ERROR},
  { "al",  ACTION_LOG_ALL, ACTION_LOG_ALL & ~ACTION_ERROR}, 
  { "yes", ACTION_LOG_ALL, ACTION_LOG_ALL & ~ACTION_ERROR}, 
  { "no",  ACTION_LOG_NONE | ACTION_ERROR, ACTION_LOG_NONE},
  { "of",  ACTION_LOG_NONE | ACTION_ERROR, ACTION_LOG_NONE},
  { "de",  ACTION_DELIVER  | ACTION_ERROR, ACTION_DELIVER},
  { "fo",  ACTION_FORWARD  | ACTION_ERROR, ACTION_FORWARD},
  { "bo",  ACTION_BOUNCE   | ACTION_ERROR, ACTION_BOUNCE},
  { "dr",  ACTION_DROP     | ACTION_ERROR, ACTION_DROP },
  { "er",  ACTION_ERROR    | ACTION_ERROR, ACTION_ERROR},
  { NULL, },
};


char *readlogtype(struct fstate *state, char *s) {
  int mode = LOGTYPEMODE_SET;

  while(*s) {
    while(*s) {
      switch(*s) {
      case '+':
	mode = LOGTYPEMODE_PLUS;  s++;break;
      case '-':
	mode = LOGTYPEMODE_MINUS; s++;break;
      case '=':
	mode = LOGTYPEMODE_SET;   s++;break;
      }
      if (isalpha((int)*s))
	break;
      s++;
    }
    if (*s) {
      char *word = s;
      while(isalpha((int)*s)) {
	s++;
      }
      if (s>word) {
	struct logtype *lt = logtypes;
	int done = 0;
	while(lt->key && done == 0) {
	  int len = strlen(lt->key);
	  if (len <= s-word && strncasecmp(lt->key, word, len)==0) {
	    switch(mode) {
	    case LOGTYPEMODE_SET:
	      state->action_log = 0;
	      mode = LOGTYPEMODE_PLUS;
	      /* FALLTHROUGH */
	    case LOGTYPEMODE_PLUS:
	      state->action_log |= lt->plus;
	      break;
	    case LOGTYPEMODE_MINUS:
	      state->action_log &= ~lt->minus;
	      break;
	    }
	    done = 1;
	  }
	  lt++;
	}
	if (!done) {
	  state->action_log = ACTION_LOG_ALL;
	  return word;
	}
      }
    }
  }


  /*  fprintf(stderr, "action_log=%s%s%s%s%s\n",
	  state->action_log & ACTION_DELIVER ? "deliver, ": "",
	  state->action_log & ACTION_BOUNCE ? "bounce, ": "",
	  state->action_log & ACTION_DROP ? "drop, ": "",
	  state->action_log & ACTION_FORWARD ? "forward, ": "",
	  state->action_log & ACTION_ERROR ? "error, ": ""
	  ); */

  return NULL;
}

static char *tagcmp(char **buf, char *tag) {
  char *xtag=*buf;
  while(**buf && !isblank((int)**buf))
    (*buf)++;
  if (**buf)
    *(*buf)++=0;
  if (strcasecmp(xtag, tag)==0) {
    *tag=0;
    skipblanks(*buf);
    /*    return **buf ? *buf : NULL; */
    return *buf; 
  }
  return NULL;
}


static char *checkbody(struct fstate *state, char **buf, int len) {
  char *xbuf=*buf;
  char *action = NULL;
  int ret;
  int ok = 1;
  RE *re;

  skipblanks(xbuf);
  if (*xbuf == '!') {
    ok = 0;
    xbuf ++;
  }

  re=re_find(&xbuf);
  if (valid_re(re) == -1) 
    return NULL;

  ret=bodchk(state, re, len);

  if (ret==ok)
    action = xbuf;

  re->close(re);
  *buf = xbuf;
  return action;
}

static char *hdrreg(struct fstate *state, char **buf,
		    struct headers *hdrlist) {
  char *xbuf=*buf;
  char *action=NULL;
  RE *re = NULL;
  int sense=RE_MATCH;
  int ret;

  while(*xbuf && !isblank((int)*xbuf))  (xbuf)++;
  if (*xbuf) *xbuf++=0;
  skipblanks(xbuf);
 
  if (*xbuf == '!') {
    sense = RE_NOTMATCH;
    xbuf ++;
  }

  if (xbuf[0] == '$' && !isblank((int)xbuf[1])) {
    re = special_re(state, &xbuf);
  }
  if (re == NULL) {
    re = re_find(&xbuf);
  }

  if (valid_re(re) == -1) 
    return NULL;

  if (**buf == '&' && strchr(*buf, ':') == NULL) {
    ret = modchk(state, re, *buf+1, sense);
  } else {
    ret = regchk(re, *buf, sense, hdrlist,
		 state->in_module ? HEADER_OTHER : state->hlevel);
  }

  if (ret == 1)
    action = xbuf;

  re->close(re);

  *buf = xbuf;
  return action;
}

/*
 * user filter parsing
 */

int filterfile_cmd(struct fstate *state, char *file, struct cmd *cmds, 
		   int (*cmd_unknown)(struct fstate *state, char *, char *)){

  FILE *ffilter;
  int final=FILTER_CONTINUE;

  if (file!= NULL && (ffilter=fopen(file, "r"))!=NULL) {
    char buf[BUFMAX];
    char buf_debug[BUFMAX];

    while(final == FILTER_CONTINUE && 
	  fgets(buf, BUFMAX-1, ffilter)!=NULL) {
      char *action=NULL;
      char *xbuf;
      char *ebuf;
      int tagok=(state->tag[0]==0);

#ifdef PARSER
      tagok=1;
      parser_nextline(buf);
#endif
      /* printf("tagok=%d, line==%s\n", tagok, buf); */

      ebuf=strchr(buf, '\n');
      if (ebuf==NULL) {
	if (feof(ffilter))
	  ebuf = buf + strlen(buf);
	else {
	  seterr(state, 452, "4.7.0", "filter line too big");
	  final = FILTER_ERROR;
	  break;
	}
      } else {
	*ebuf--=0; /* last character of the line */
      }

      xbuf=buf;
      skipblanks(xbuf);                    /* skip starting blanks */

      while(ebuf > xbuf && isblank(*ebuf)) /* remove trailing blanks */
      	*ebuf--=0;

      if (state->filter_debug) {
	strcpy(buf_debug, xbuf);
      } 

      switch (*xbuf++) {
      case '\n':
      case '\r':
      case '\0':
      case '#':
	continue;
	
      case ':':
	action=tagcmp(&xbuf, state->tag);
	if (action && *action == 0) {
	  action = NULL;
	  fdebug(state, "+ %s\n", buf_debug);
	}
#ifdef PARSER
	if (*xbuf==0) xbuf = strdup("drop"); /* for an empty action */
#endif
	break;

      case '-': case '+': case '0': case '1': case '2': case '3':
      case '4': case '5': case '6': case '7': case '8': case '9':
	if (!tagok) continue;
	{
	  int len = strtol(xbuf-1, &xbuf, 10);
	  action=checkbody(state, &xbuf, len);
	}
	break;
      case '=': case '<': case '>':
	if (!tagok) continue;
	xbuf --;
	/* FALLTHROUGH */
      case '$':
	if (!tagok) continue;
	{
	  int ok = test_scalar(state, &xbuf);
	  if (ok && *xbuf && isblank((int)*xbuf)) {
	    action = xbuf;
	  } 
	}
	break;

      default:
	if (!tagok) continue;
	xbuf--;
	action=hdrreg(state, &xbuf, state->hdrlist);
	break;
      }
#ifdef PARSER
      action=xbuf;
#endif
      if (action) {
	char *name;
	struct cmd *xcmd=cmds;
	size_t namelen;

	fdebug(state, "+ %s\n", buf_debug);

	xbuf=action;
	skipblanks(xbuf);
	name=xbuf;
	while(*xbuf && !isspace((int)*xbuf)) { 
	  *xbuf = tolower((int)*xbuf); /* lower the command case */
	  xbuf ++;
	} 
	namelen = xbuf - name;

	if (*xbuf) *xbuf++=0;

	while(xcmd->name) {
	  if (namelen == xcmd->len && 
	      memcmp(xcmd->name, name, namelen)==0) {   /* seek the command */
	    /* remove starting and ending blanks */
	    while(ebuf>=xbuf && isblank((int)*ebuf)) *ebuf--=0;
	    skipblanks(xbuf);
	    final = xcmd->f(state, xbuf);  /* run the command */
	    break;                         /* exit loop */
	  }
	  xcmd++;
	}
	if (!xcmd->name && cmd_unknown)
	  final = cmd_unknown(state, name, xbuf);
      }
    }
    fclose(ffilter);
  }

  return final; 
}

int filterfile(struct fstate *state, char *filename) {
  return filterfile_cmd(state, filename, cmdlist, unknown_cmd);
}

#ifndef PARSER
char *filter (struct fstate *state) {
  char **file = state->filtername.argv;
  int final = FILTER_CONTINUE;
  while(*file && final != FILTER_STOP) {
    char *filename = *file++;
    char *prefix = strchr(filename, ':');
    resetstate(state, STATE_FILTER);
    if (prefix) {
      struct {
	char *s;
	int l;
	int (*filter)(struct fstate *, char *); 
      } static pf[] = {
	{ "file",  sizeof("file")-1 , filterfile }, 
#if WITH_PERL
	{ "perl",  sizeof("perl")-1 , filterperl },
#endif
#if WITH_LIBSIEVE
	{ "sieve", sizeof("sieve")-1, filtersieve },
#endif      
	{ NULL, -1, NULL},
      }, *pfi = pf;
      while(pfi->s) {
	if (prefix - filename == pfi->l &&
	    strncmp(filename, pfi->s, pfi->l) == 0) {
	  final = pfi->filter(state, prefix + 1);
	  break;
	}
	pfi++;
      }
      if (pfi->s == NULL) 
	syslog(LOG_ERR, "unknown filter type %s", filename);
    } else {
      final = filterfile(state, filename);
    }
  }
  if (!state->fatal) {
    if (final != FILTER_STOP) 
      default_cmd(state);
    flush_action(state);
    if (state->filter_debug) {
      debug_close(state, state->filter_debug);
      fclose(state->filter_debug);
      state->filter_debug = NULL;
    }
  }
  return state->error[0] ? state->error : NULL ;
}
#endif

==> lmtpd-0.9.24.tar.gz: lmtpd-0.9.24/filter.c <==
/*
   Copyright 2000, 2001, 2002, 2003 Laurent Wacrenier

   This file is part of PLL lmtpd.

   PLL lmtpd is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   PLL lmtpd is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with PLL lmtpd; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "config.h"

static char const rcsid[] UNUSED =
"$Id: filter.c,v 1.67 2004/10/26 09:59:20 lwa Exp $";



#include <sys/types.h>
#include <syslog.h>

#include <limits.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <strings.h>
#include <string.h>

#include "lmtp.h"

#define BUFMAX 1024

#if !HAVE_ISBLANK
int isblank(int c) {
  return c == ' ' || c == '\t';
}
#endif

#define skipblanks(xbuf) while(isblank((int)*(xbuf))) (xbuf)++

#define CMD(x) { #x, sizeof(#x)-1, cmd_##x, }
#define CMD_END { NULL, 0, NULL, }

struct cmd cmdlist[] = {
  /* { "deliver", sizeof("deliver")-1, cmd_deliver}, */
  CMD(deliver),
  CMD(drop),
  CMD(bounce),
  CMD(forward),
  CMD(forward2),
  CMD(skipto),
  CMD(score),
  CMD(trash),
  CMD(vacation),
  CMD(logrotate),
  CMD(trace),
  CMD(timezone),
  CMD(mod),
  CMD(add),
  CMD(blacklist),
  CMD(extension),
  CMD(stow),
  CMD(log),
  CMD(syslog),
  CMD(deliver2),
  CMD(default),
  CMD(antivirus),
  CMD(spamassassin),
  CMD(header_set),
  CMD(archive),
  CMD(archive2),
  CMD(moved),
  CMD(pipe),
  CMD(debug),
  CMD(stop),
  CMD_END,
};

struct cmd mod_cmdlist[] = {
  CMD(score),
  CMD(message),
  CMD(ok),
  CMD(skipto),
  CMD_END,
};

#define LOGTYPEMODE_SET   '='
#define LOGTYPEMODE_PLUS  '+'
#define LOGTYPEMODE_MINUS '-'

struct logtype {
  char *key;
  unsigned char plus;
  unsigned char minus;
} logtypes[] = {
  { "on",  ACTION_LOG_ALL, ACTION_LOG_ALL & ~ACTION_ERROR},
  { "al",  ACTION_LOG_ALL, ACTION_LOG_ALL & ~ACTION_ERROR}, 
  { "yes", ACTION_LOG_ALL, ACTION_LOG_ALL & ~ACTION_ERROR}, 
  { "no",  ACTION_LOG_NONE | ACTION_ERROR, ACTION_LOG_NONE},
  { "of",  ACTION_LOG_NONE | ACTION_ERROR, ACTION_LOG_NONE},
  { "de",  ACTION_DELIVER  | ACTION_ERROR, ACTION_DELIVER},
  { "fo",  ACTION_FORWARD  | ACTION_ERROR, ACTION_FORWARD},
  { "bo",  ACTION_BOUNCE   | ACTION_ERROR, ACTION_BOUNCE},
  { "dr",  ACTION_DROP     | ACTION_ERROR, ACTION_DROP },
  { "er",  ACTION_ERROR    | ACTION_ERROR, ACTION_ERROR},
  { NULL, },
};


char *readlogtype(struct fstate *state, char *s) {
  int mode = LOGTYPEMODE_SET;

  while(*s) {
    while(*s) {
      switch(*s) {
      case '+':
	mode = LOGTYPEMODE_PLUS;  s++;break;
      case '-':
	mode = LOGTYPEMODE_MINUS; s++;break;
      case '=':
	mode = LOGTYPEMODE_SET;   s++;break;
      }
      if (isalpha((int)*s))
	break;
      s++;
    }
    if (*s) {
      char *word = s;
      while(isalpha((int)*s)) {
	s++;
      }
      if (s>word) {
	struct logtype *lt = logtypes;
	int done = 0;
	while(lt->key && done == 0) {
	  int len = strlen(lt->key);
	  if (len <= s-word && strncasecmp(lt->key, word, len)==0) {
	    switch(mode) {
	    case LOGTYPEMODE_SET:
	      state->action_log = 0;
	      mode = LOGTYPEMODE_PLUS;
	      /* FALLTHROUGH */
	    case LOGTYPEMODE_PLUS:
	      state->action_log |= lt->plus;
	      break;
	    case LOGTYPEMODE_MINUS:
	      state->action_log &= ~lt->minus;
	      break;
	    }
	    done = 1;
	  }
	  lt++;
	}
	if (!done) {
	  state->action_log = ACTION_LOG_ALL;
	  return word;
	}
      }
    }
  }


  /*  fprintf(stderr, "action_log=%s%s%s%s%s\n",
	  state->action_log & ACTION_DELIVER ? "deliver, ": "",
	  state->action_log & ACTION_BOUNCE ? "bounce, ": "",
	  state->action_log & ACTION_DROP ? "drop, ": "",
	  state->action_log & ACTION_FORWARD ? "forward, ": "",
	  state->action_log & ACTION_ERROR ? "error, ": ""
	  ); */

  return NULL;
}

static char *tagcmp(char **buf, char *tag) {
  char *xtag=*buf;
  while(**buf && !isblank((int)**buf))
    (*buf)++;
  if (**buf)
    *(*buf)++=0;
  if (strcasecmp(xtag, tag)==0) {
    *tag=0;
    skipblanks(*buf);
    /*    return **buf ? *buf : NULL; */
    return *buf; 
  }
  return NULL;
}


static char *checkbody(struct fstate *state, char **buf, int len) {
  char *xbuf=*buf;
  char *action = NULL;
  int ret;
  int ok = 1;
  RE *re;

  skipblanks(xbuf);
  if (*xbuf == '!') {
    ok = 0;
    xbuf ++;
  }

  re=re_find(&xbuf);
  if (valid_re(re) == -1) 
    return NULL;

  ret=bodchk(state, re, len);

  if (ret==ok)
    action = xbuf;

  re->close(re);
  *buf = xbuf;
  return action;
}

static char *hdrreg(struct fstate *state, char **buf,
		    struct headers *hdrlist) {
  char *xbuf=*buf;
  char *action=NULL;
  RE *re = NULL;
  int sense=RE_MATCH;
  int ret;

  while(*xbuf && !isblank((int)*xbuf))  (xbuf)++;
  if (*xbuf) *xbuf++=0;
  skipblanks(xbuf);
 
  if (*xbuf == '!') {
    sense = RE_NOTMATCH;
    xbuf ++;
  }

  if (xbuf[0] == '$' && !isblank((int)xbuf[1])) {
    re = special_re(state, &xbuf);
  }
  if (re == NULL) {
    re = re_find(&xbuf);
  }

  if (valid_re(re) == -1) 
    return NULL;

  if (**buf == '&' && strchr(*buf, ':') == NULL) {
    ret = modchk(state, re, *buf+1, sense);
  } else {
    ret = regchk(re, *buf, sense, hdrlist,
		 state->in_module ? HEADER_OTHER : state->hlevel);
  }

  if (ret == 1)
    action = xbuf;

  re->close(re);

  *buf = xbuf;
  return action;
}

/*
 * user filter parsing
 */

int filterfile_cmd(struct fstate *state, char *file, struct cmd *cmds, 
		   int (*cmd_unknown)(struct fstate *state, char *, char *)){

  FILE *ffilter;
  int final=FILTER_CONTINUE;

  if (file!= NULL && (ffilter=fopen(file, "r"))!=NULL) {
    char buf[BUFMAX];
    char buf_debug[BUFMAX];

#ifndef PARSER
    state->charset = find_relist(state->file_charset_re, file);
#endif      

    while(final == FILTER_CONTINUE && 
	  fgets(buf, BUFMAX-1, ffilter)!=NULL) {
      char *action=NULL;
      char *xbuf;
      char *ebuf;
      int tagok=(state->tag[0]==0);

#ifdef PARSER
      tagok=1;
      parser_nextline(buf);
#endif
      /* printf("tagok=%d, line==%s\n", tagok, buf); */

      ebuf=strchr(buf, '\n');
      if (ebuf==NULL) {
	if (feof(ffilter))
	  ebuf = buf + strlen(buf);
	else {
	  seterr(state, 452, "4.7.0", "filter line too big");
	  final = FILTER_ERROR;
	  break;
	}
      } else {
	*ebuf--=0; /* last character of the line */
      }

      xbuf=buf;
      skipblanks(xbuf);                    /* skip starting blanks */

      while(ebuf > xbuf && isblank(*ebuf)) /* remove trailing blanks */
      	*ebuf--=0;

      if (state->filter_debug) {
	strcpy(buf_debug, xbuf);
      } 

      switch (*xbuf++) {
      case '\n':
      case '\r':
      case '\0':
      case '#':
	continue;
	
      case ':':
	action=tagcmp(&xbuf, state->tag);
	if (action && *action == 0) {
	  action = NULL;
	  fdebug(state, "+ %s\n", buf_debug);
	}
#ifdef PARSER
	if (*xbuf==0) xbuf = strdup("drop"); /* for an empty action */
#endif
	break;

      case '-': case '+': case '0': case '1': case '2': case '3':
      case '4': case '5': case '6': case '7': case '8': case '9':
	if (!tagok) continue;
	{
	  int len = strtol(xbuf-1, &xbuf, 10);
	  action=checkbody(state, &xbuf, len);
	}
	break;
      case '=': case '<': case '>':
	if (!tagok) continue;
	xbuf --;
	/* FALLTHROUGH */
      case '$':
	if (!tagok) continue;
	{
	  int ok = test_scalar(state, &xbuf);
	  if (ok && *xbuf && isblank((int)*xbuf)) {
	    action = xbuf;
	  } 
	}
	break;

      default:
	if (!tagok) continue;
	xbuf--;
	action=hdrreg(state, &xbuf, state->hdrlist);
	break;
      }
#ifdef PARSER
      action=xbuf;
#endif
      if (action) {
	char *name;
	struct cmd *xcmd=cmds;
	size_t namelen;

	fdebug(state, "+ %s\n", buf_debug);

	xbuf=action;
	skipblanks(xbuf);
	name=xbuf;
	while(*xbuf && !isspace((int)*xbuf)) { 
	  *xbuf = tolower((int)*xbuf); /* lower the command case */
	  xbuf ++;
	} 
	namelen = xbuf - name;

	if (*xbuf) *xbuf++=0;

	while(xcmd->name) {
	  if (namelen == xcmd->len && 
	      memcmp(xcmd->name, name, namelen)==0) {   /* seek the command */
	    /* remove starting and ending blanks */
	    while(ebuf>=xbuf && isblank((int)*ebuf)) *ebuf--=0;
	    skipblanks(xbuf);
	    final = xcmd->f(state, xbuf);  /* run the command */
	    break;                         /* exit loop */
	  }
	  xcmd++;
	}
	if (!xcmd->name && cmd_unknown)
	  final = cmd_unknown(state, name, xbuf);
      }
    }
    fclose(ffilter);
  }

  return final; 
}

int filterfile(struct fstate *state, char *filename) {
  return filterfile_cmd(state, filename, cmdlist, unknown_cmd);
}

#ifndef PARSER
char *filter (struct fstate *state) {
  char **file = state->filtername.argv;
  int final = FILTER_CONTINUE;
  while(*file && final != FILTER_STOP) {
    char *filename = *file++;
    char *prefix = strchr(filename, ':');
    resetstate(state, STATE_FILTER);
    if (prefix) {
      struct {
	char *s;
	int l;
	int (*filter)(struct fstate *, char *); 
      } static pf[] = {
	{ "file",  sizeof("file")-1 , filterfile }, 
#if WITH_PERL
	{ "perl",  sizeof("perl")-1 , filterperl },
#endif
#if WITH_LIBSIEVE
	{ "sieve", sizeof("sieve")-1, filtersieve },
#endif      
	{ NULL, -1, NULL},
      }, *pfi = pf;
      while(pfi->s) {
	if (prefix - filename == pfi->l &&
	    strncmp(filename, pfi->s, pfi->l) == 0) {
	  final = pfi->filter(state, prefix + 1);
	  break;
	}
	pfi++;
      }
      if (pfi->s == NULL) 
	syslog(LOG_ERR, "unknown filter type %s", filename);
    } else {
      final = filterfile(state, filename);
    }
  }
  if (!state->fatal) {
    if (final != FILTER_STOP) 
      default_cmd(state);
    flush_action(state);
    if (state->filter_debug) {
      debug_close(state, state->filter_debug);
      fclose(state->filter_debug);
      state->filter_debug = NULL;
    }
  }
  return state->error[0] ? state->error : NULL ;
}
#endif
